/*

// File:    A-Star Pathfinding.dm
// Library: Xirre.Pathfinding
// Author:  Xirre

Example of use:
//You may want to organize your z cords in to lists of pathfinders. See the demo.
var/Pathfinder/pathfinder = new/Pathfinder(world.maxx, world.maxy, src.z)

var/list/Vector2/path = pf.FindPath(new/Point(src.x, src.y), new/Point(_x, _y))

//Then ;oop through path and move your mob with each of path's values.


*/

#define FLOAT_MAX 3.402823e38

Point
	parent_type = /obj

	New(xLoc, yLoc)
		src.x = xLoc
		src.y = yLoc

Vector2
	var/x
	var/y

	New(_x, _y)
		x = _x
		y = _y

SearchNode
	var
		Point/pos
		dense
		list/SearchNode/neighbors
		SearchNode/parent
		inOpenList
		inClosedList
		distToGoal
		distTraveled

Pathfinder
	var
		list
			SearchNode
				searchNodes[][]
				openList[]
				closedList[]

		levelWidth
		levelHeight
		map

	New(_width, _height, _map)
		if(!_width || !_height || !_map)
			del src

		levelWidth = _width
		levelHeight = _height
		map = _map

		InitializeSearchNodes()

	proc
		Heuristic(Point/point1, Point/point2)
			return abs(point1.x - point2.x) + abs(point1.y - point2.y) // TODO: Try with get_dist().

		ResetSearchNodes()
			openList = new/list()
			closedList = new/list()

			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					var/SearchNode/node = searchNodes[x][y]

					if(node == null)
						continue

					node.inOpenList = 0
					node.inClosedList = 0

					node.distTraveled = FLOAT_MAX
					node.distToGoal = FLOAT_MAX

		FindBestNode()
			var/SearchNode/currentTile = openList[1]

			var/smallestDistToGoal = FLOAT_MAX

			for(var/i = 1; i <= length(openList); i++)
				var/SearchNode/oL = openList[i]
				if(oL.distToGoal < smallestDistToGoal)
					currentTile = openList[i]
					smallestDistToGoal = currentTile.distToGoal

			return currentTile

		FindFinalPath(SearchNode/startNode, SearchNode/endNode)
			closedList.Add(endNode)

			var/SearchNode/parentTile = endNode.parent

			while(parentTile != startNode)
				closedList.Add(parentTile)
				parentTile = parentTile.parent

			var/list/Vector2/finalPath = list()

			for(var/i = length(closedList); i >= 1; i--)
				var/SearchNode/cL = closedList[i]
				finalPath.Add(new/Vector2(cL.pos.x, cL.pos.y))

			return finalPath

		FindPath(Point/startPoint, Point/endPoint)

			if(startPoint == endPoint)
				return null

			ResetSearchNodes()

			var/SearchNode/startNode = searchNodes[startPoint.x][startPoint.y]
			var/SearchNode/endNode = searchNodes[endPoint.x][endPoint.y]

			startNode.inOpenList = 1

			startNode.distToGoal = Heuristic(startPoint, endPoint)
			startNode.distTraveled = 0

			openList.Add(startNode)

			while(length(openList) > 0)
				var/SearchNode/currentNode = FindBestNode()

				if(currentNode == null)
					break

				if(currentNode == endNode)
					return FindFinalPath(startNode, endNode)

				for(var/i = 1; i <= length(currentNode.neighbors); i++)
					var/SearchNode/neighbor = currentNode.neighbors[i]

					if(neighbor == null || !neighbor.dense)
						continue

					var/distTraveled = currentNode.distTraveled + 1

					var/heuristic = Heuristic(neighbor.pos, endPoint)

					if(!neighbor.inOpenList && !neighbor.inClosedList)
						neighbor.distTraveled = distTraveled
						neighbor.distToGoal = distTraveled + heuristic
						neighbor.parent = currentNode
						neighbor.inOpenList = 1
						openList.Add(neighbor)

					else if(neighbor.inOpenList || neighbor.inClosedList)
						if(neighbor.distTraveled > distTraveled)
							neighbor.distTraveled = distTraveled
							neighbor.distToGoal = distTraveled + heuristic

							neighbor.parent = currentNode

				openList.Remove(currentNode)
				currentNode.inClosedList = 1

			return null

		InitializeSearchNodes()
			searchNodes = new/list(levelWidth, levelHeight)

			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					var/SearchNode/node = new/SearchNode()
					node.pos = locate(x, y, map)

					node.dense = node.pos.density == 0

					if(node.dense)
						node.neighbors = new/list(4)
						searchNodes[x][y] = node

			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					var/SearchNode/node = searchNodes[x][y]

					if(node == null || !node.dense)
						continue

					var/Point/neighbors = list(
						new/Point(x, y - 1),
						new/Point(x, y + 1),
						new/Point(x - 1, y),
						new/Point(x + 1, y)
					)

					for(var/i = 1; i <= length(neighbors); i++)
						var/Point/position = neighbors[i]

						if(position.x <= 0 || position.x > levelWidth || position.y <= 0 || position.y > levelHeight)
							continue

						var/SearchNode/neighbor = searchNodes[position.x][position.y]

						if(neighbor == null || !neighbor.dense)
							continue

						node.neighbors[i] = neighbor