/*

// The A* Algorithm Reminder (http://xnatd.blogspot.com/2011/08/pathfinding-tutorial-part-2.html)

The G Value - Each node will have an associated G value, this is how far
we would have to travel to get from the starting node to this node.

The H Function (The Heuristic Function) - This is a method that will help
us to estimate how far we would have to move to get from one node to another
node "as the crow flies". The method will not take into account any obstacles
which is why it will only be an estimate.

The F Value - Each node will also have an associated F value. This is a rough
estimate of how far we would have to travel to get from the starting node to
the end node if the path crosses this node. To work out an F value we can do
the following : F = G + H - it is basically how far we will have already had
to travel to get to this node add a rough estimate of how much further we need
to travel.

The Open List - This is basically a big list of all the nodes that are being
considered to be included in the final path.

The Closed List - This is a list of all of the nodes that have already have been
looked at, we will use this to make sure we dont consider each node more than once.

The Parent Reference - Each node will have a reference to the node that added this
node to the Open List. When the algorithm has finished, we will use the reference
to the nodes Parent to trace the final path from the end node to the start node
( I will cover this in more detail later ).


#define FLOAT_MAX 3.402823e38 // float.MaxValue in C#.

Point
	parent_type = /obj

	New(xLoc, yLoc)
		src.x = xLoc
		src.y = yLoc

Vector2
	var/x
	var/y

	New(_x, _y)
		x = _x
		y = _y

SearchNode
	var
		Point/pos // Location on map.
		dense // Can be walked on.
		list/SearchNode/neighbors // Nodes surrounding this tile (up, down, left, right).
		SearchNode/parent // A reference to the node that transferred this node to the open list. This will be used to trace our path back from the goal node to the start node.
		inOpenList // An easy way to check if this node is in the open list.
		inClosedList // An easy way to check if this node is in the closed list.
		distToGoal // The approximate distance from start node to the goal node if the path goes through this node. (F)
		distTraveled // Distance traveled from the start point. (G)

Pathfinder
	var
		list
			SearchNode
				searchNodes[][] // Stores an array of the walkable search nodes.
				openList[] // Holds search nodes that are available to search.
				closedList[] // Holds the nodes that have already been searched.

		levelWidth // maxx
		levelHeight // maxy
		map // Basically the current Z coordinate.

	// Constructor
	New(_width, _height, _map)
		if(!_width || !_height || !_map)
			del src

		levelWidth = _width
		levelHeight = _height
		map = _map

		InitializeSearchNodes()

	proc
		//Returns an estimate of the distance between two points. (H)
		Heuristic(Point/point1, Point/point2)
			return abs(point1.x - point2.x) + abs(point1.y - point2.y) // TODO: Try with get_dist().


		// Resets the state of the search nodes.
		ResetSearchNodes()
			openList = new/list()
			closedList = new/list()

			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					var/SearchNode/node = searchNodes[x][y]

					if(node == null)
						continue

					node.inOpenList = 0
					node.inClosedList = 0

					node.distTraveled = FLOAT_MAX
					node.distToGoal = FLOAT_MAX

		// Returns the node with the smallest distance to goal.
		FindBestNode()
			var/SearchNode/currentTile = openList[1]

			var/smallestDistToGoal = FLOAT_MAX

			// Find the closest node to the goal.
			for(var/i = 1; i <= length(openList); i++)
				var/SearchNode/oL = openList[i]
				if(oL.distToGoal < smallestDistToGoal)
					currentTile = openList[i]
					smallestDistToGoal = currentTile.distToGoal

			return currentTile

		// Use the parent field of the search nodes to trace a path from the end node to the start node.
		FindFinalPath(SearchNode/startNode, SearchNode/endNode)
			closedList.Add(endNode)

			var/SearchNode/parentTile = endNode.parent

			// Trace back through the nodes using the parent fields to find the best path.
			while(parentTile != startNode)
				closedList.Add(parentTile)
				parentTile = parentTile.parent

			var/list/Vector2/finalPath = list()

			// Reverse the path and transform into world space.
			for(var/i = length(closedList); i >= 1; i--)
				var/SearchNode/cL = closedList[i]
				finalPath.Add(new/Vector2(cL.pos.x, cL.pos.y))

			return finalPath

		// Finds the optimal path from one point to another.
		FindPath(Point/startPoint, Point/endPoint)
			// Only try to find a path if the start and end points are different.
			if(startPoint == endPoint)
				return null

			/////////////////////////////////////////////////////////////////////
			// Step 1 : Clear the Open and Closed Lists and reset each nodes F
			//          and G values in case they are still set from the last
			//          time we tried to find a path.
			/////////////////////////////////////////////////////////////////////
			ResetSearchNodes()

			// Store references to the start and end nodes for convenience.
			var/SearchNode/startNode = searchNodes[startPoint.x][startPoint.y]
			var/SearchNode/endNode = searchNodes[endPoint.x][endPoint.y]

			/////////////////////////////////////////////////////////////////////
			// Step 2 : Set the start nodes G value to 0 and its F value to the
			//          estimated distance between the start node and goal node
			//          (this is where our H function comes in) and add it to the
			//          Open List.
			/////////////////////////////////////////////////////////////////////
			startNode.inOpenList = 1

			startNode.distToGoal = Heuristic(startPoint, endPoint)
			startNode.distTraveled = 0

			openList.Add(startNode)

			/////////////////////////////////////////////////////////////////////
			// Setp 3 : While there are still nodes to look at in the Open list :
			/////////////////////////////////////////////////////////////////////
			while(length(openList) > 0)
				/////////////////////////////////////////////////////////////////
				// a) : Loop through the Open List and find the node that
				//      has the smallest F value.
				/////////////////////////////////////////////////////////////////
				var/SearchNode/currentNode = FindBestNode()

				/////////////////////////////////////////////////////////////////
				// b) : If the Open List empty or no node can be found,
				//      no path can be found so the algorithm terminates.
				/////////////////////////////////////////////////////////////////
				if(currentNode == null)
					break

				/////////////////////////////////////////////////////////////////
				// c) : If the Active Node is the goal node, we will
				//      find and return the final path.
				/////////////////////////////////////////////////////////////////
				if(currentNode == endNode)
					// Trace our path back to the start.
					return FindFinalPath(startNode, endNode)

				/////////////////////////////////////////////////////////////////
				// d) : Else, for each of the Active Nodes neighbours :
				/////////////////////////////////////////////////////////////////
				for(var/i = 1; i <= length(currentNode.neighbors); i++)
					var/SearchNode/neighbor = currentNode.neighbors[i]

					//////////////////////////////////////////////////
					// i) : Make sure that the neighbouring node can
					//      be walked across.
					//////////////////////////////////////////////////
					if(neighbor == null || !neighbor.dense)
						continue

					//////////////////////////////////////////////////
					// ii) Calculate a new G value for the neighbouring node.
					//////////////////////////////////////////////////
					var/distTraveled = currentNode.distTraveled + 1

					// An estimate of the distance from this node to the end node.
					var/heuristic = Heuristic(neighbor.pos, endPoint)

					//////////////////////////////////////////////////
					// iii) If the neighbouring node is not in either the Open
					//      List or the Closed List :
					//////////////////////////////////////////////////
					if(!neighbor.inOpenList && !neighbor.inClosedList)
						// (1) Set the neighbouring nodes G value to the G value
						//     we just calculated.
						neighbor.distTraveled = distTraveled
						// (2) Set the neighbouring nodes F value to the new G value +
						//     the estimated distance between the neighbouring node and
						//     goal node.
						neighbor.distToGoal = distTraveled + heuristic
						// (3) Set the neighbouring nodes Parent property to point at the Active
						//     Node.
						neighbor.parent = currentNode
						// (4) Add the neighbouring node to the Open List.
						neighbor.inOpenList = 1
						openList.Add(neighbor)

					//////////////////////////////////////////////////
					// iv) Else if the neighbouring node is in either the Open
					//     List or the Closed List :
					//////////////////////////////////////////////////
					else if(neighbor.inOpenList || neighbor.inClosedList)
						// (1) If our new G value is less than the neighbouring
						//     nodes G value, we basically do exactly the same
						//     steps as if the nodes are not in the Open and
						//     Closed Lists except we do not need to add this node
						//     the Open List again.
						if(neighbor.distTraveled > distTraveled)
							neighbor.distTraveled = distTraveled
							neighbor.distToGoal = distTraveled + heuristic

							neighbor.parent = currentNode

				/////////////////////////////////////////////////////////////////
				// e) Remove the Active Node from the Open List and add it to the
				//    Closed List
				/////////////////////////////////////////////////////////////////
				openList.Remove(currentNode)
				currentNode.inClosedList = 1

			// No path could be found.
			return null

		// Split the map in to a grid of nodes.
		InitializeSearchNodes()
			searchNodes = new/list(levelWidth, levelHeight)

			// For every tile on the map, create a search node for it.
			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					// Create a search node to represent this tile.
					var/SearchNode/node = new/SearchNode()
					node.pos = locate(x, y, map)

					// Our enemies can only walk on grass tiles.
					node.dense = node.pos.density == 0

					// Only store nodes that can be walked on.
					if(node.dense)
						node.neighbors = new/list(4)
						searchNodes[x][y] = node

			// Connect each search node to its neighbors.
			for(var/x = 1; x <= levelWidth; x++)
				for(var/y = 1; y <= levelHeight; y++)
					var/SearchNode/node = searchNodes[x][y]

					// Only look at the nodes our enemies can walk on.
					if(node == null || !node.dense)
						continue

					var/Point/neighbors = list(
						new/Point(x, y - 1), // The node above the current node.
						new/Point(x, y + 1), // The node below the current node.
						new/Point(x - 1, y), // The node left of the current node.
						new/Point(x + 1, y) // The node right of the current node.
					)

					// Loop through each of the possible neighbors.
					for(var/i = 1; i <= length(neighbors); i++)
						var/Point/position = neighbors[i]

						// Make sure this neighbor is part of the map.
						if(position.x <= 0 || position.x > levelWidth || position.y <= 0 || position.y > levelHeight)
							continue
						var/SearchNode/neighbor = searchNodes[position.x][position.y]

						// Only bother keeping a reference to nodes that can be walked on.
						if(neighbor == null || !neighbor.dense)
							continue

						// Store a reference to the neighbor.
						node.neighbors[i] = neighbor
*/